package ru.dbotthepony.kstarbound.world.entities

import com.google.gson.JsonArray
import ru.dbotthepony.kstarbound.Starbound
import ru.dbotthepony.kstarbound.defs.actor.ActorMovementModifiers
import ru.dbotthepony.kstarbound.defs.ActorMovementParameters
import ru.dbotthepony.kstarbound.math.AABB
import ru.dbotthepony.kstarbound.math.vector.Vector2d
import ru.dbotthepony.kstarbound.util.supplyAsync
import ru.dbotthepony.kstarbound.world.Direction
import ru.dbotthepony.kstarbound.world.entities.api.ScriptedEntity
import ru.dbotthepony.kstarbound.world.physics.CollisionType
import ru.dbotthepony.kstarbound.world.physics.Poly
import java.util.concurrent.CompletableFuture
import java.util.function.Predicate
import kotlin.math.PI
import kotlin.math.min
import kotlin.math.roundToInt

class PathController(val controller: ActorMovementController, var edgeTimer: Double = 0.0) {
	var startPosition: Vector2d? = null
		private set
	var endPosition: Vector2d? = null
		private set
	var controlFace: Direction? = null
		private set

	val world get() = controller.world

	fun reset() {
		edgeTimer = 0.0
		startPosition = null
		endPosition = null
		controlFace = null
		path.clear()
		pathFinder?.signalStop()
		pathFinder = null
		pathFinderResult = null
		target = null
	}

	private var pathFinder: PathFinder? = null
	private var pathFinderResult: CompletableFuture<List<PathFinder.Edge>?>? = null
	private val path = ArrayDeque<PathFinder.Edge>()
	private var target: Vector2d? = null
	var parameters = PathFinder.Parameters()

	val currentEdge: PathFinder.Edge?
		get() = path.firstOrNull()

	val currentAction: PathFinder.Action?
		get() = currentEdge?.action

	private fun onGround(position: Vector2d, type: Predicate<CollisionType>): Boolean {
		val box = controller.computeGlobalCollisionAABB()

		return controller.world.chunkMap.anyCellSatisfies(AABB(box.mins - Vector2d.POSITIVE_Y, box.mins))
			{ _, _, cell -> type.test(cell.foreground.material.value.collisionKind) }
	}

	private fun updatePathFinding(): Boolean? {
		if (this.pathFinderResult != null && this.pathFinderResult!!.isDone) {
			val pathFinder = this.pathFinderResult!!
			this.pathFinder = null
			this.pathFinderResult = null
			val newPath = pathFinder.get()

			if (newPath == null) {
				reset()
				return false
			}

			var newEdgeTimer = 0.0
			var merged = false

			// if we have a path already, see if our paths can be merged either by splicing or fast forwarding
			if (this.path.isNotEmpty() && newPath.isNotEmpty()) {
				// try to fast-forward
				val firstNode = this.path[0]
				var ffIndex = newPath.indexOfFirst { firstNode.isSame(it) }

				if (ffIndex != -1) {
					merged = true
					this.path.clear()

					for (i in ffIndex until newPath.size) {
						this.path.add(newPath[i])
					}
				} else {
					// try to splice the new path onto the current path
					ffIndex = this.path.indexOfFirst { path[0].source == it.target }

					if (ffIndex != -1) {
						// splice the new path onto our current path up to this index
						merged = true
						newEdgeTimer = this.edgeTimer

						while (this.path.size > ffIndex) {
							this.path.removeLast()
						}

						this.path.addAll(newPath)
					}
				}
			}

			if (!merged && controller.position != this.startPosition) {
				// merging the paths failed, and the entity has moved from the path start position
				// try to bridge the gap from the current position to the new path
				val bridgePathFinder = PathFinder(this, this.startPosition ?: controller.position)
				bridgePathFinder.run(controller.actorMovementParameters.pathExploreRate?.roundToInt() ?: 100)
				val bridgePath = bridgePathFinder.result.orNull()

				if (bridgePath == null) {
					// if the gap isn't bridged in a single tick, reset and start over
					reset()
					return null
				} else {
					// concatenate the bridge path with the new path
					this.path.clear()
					this.path.addAll(bridgePath)
					this.path.addAll(newPath)
				}
			}

			if (this.path.isNotEmpty() && !validateEdge(this.path[0])) {
				// reset if the first edge is invalid
				reset()
				return false
			}

			this.edgeTimer = newEdgeTimer
			return this.path.isEmpty()
		} else {
			return null
		}
	}

	fun findPath(target: Vector2d): Boolean? {
		// reached the end of the last path and we have a new target position to move toward

		if (path.isEmpty() && controller.world.geometry.diff(this.target!!, target).lengthSquared > 0.1) {
			reset()
			this.target = target
		}

		// starting a new path, or the target position moved by more than 2 blocks
		if (this.target == null || (this.path.isEmpty() && this.pathFinder == null) || controller.world.geometry.diff(this.target!!, target).lengthSquared > 4.0) {
			var grounded = controller.isOnGround

			// if already moving on a path, collision will be disabled and we can't use MovementController::onGround() to check for ground collision
			if (this.path.isNotEmpty()) {
				grounded = onGround(controller.position) { it.isFloorCollision }
			}

			if (controller.actorMovementParameters.gravityEnabled == true && !grounded && !controller.isLiquidMovement) {
				return null
			}

			// interrupt ongoing pathfinding
			this.pathFinder?.signalStop()
			this.startPosition = controller.position
			this.target = target
			this.pathFinder = PathFinder(this, target)
			this.pathFinderResult = Starbound.EXECUTOR.supplyAsync(this.pathFinder!!)
		}

		if (this.pathFinderResult == null && this.path.isEmpty())
			return true // Reached goal

		return updatePathFinding()
	}

	private fun openDoors(bounds: AABB): Boolean {
		val entities = world.entityIndex
			.query(bounds, Predicate {
				it is ScriptedEntity && !it.isRemote
			})

		// preserve original engine behavior where it will open ALL potential doors on Actor's path
		var any = false

		for (it in entities) {
			val scripted = it as ScriptedEntity
			val result = scripted.callScript("hasCapability", "closedDoor")

			if (result.isNotEmpty() && result[0] is Boolean && result[0] as Boolean) {
				it.dispatchMessage(world.connectionID, "openDoor", JsonArray())
				any = true
			}
		}

		return any
	}

	private fun movingCollision(poly: Poly): Boolean {
		// TODO
		return false
	}

	private fun inLiquid(position: Vector2d): Boolean {
		val liquidLevel = world.chunkMap.averageLiquidLevel(controller.computeGlobalCollisionAABB() + position) ?: return false
		return liquidLevel.average >= (controller.actorMovementParameters.minimumLiquidPercentage ?: 1.0)
	}

	private fun validateEdge(edge: PathFinder.Edge): Boolean {
		val polies = controller.actorMovementParameters.standingPoly?.map({ listOf(it + edge.target.position) }, { it.map { it + edge.target.position } }) ?: listOf()

		for (poly in polies) {
			if (world.chunkMap.collide(poly, Predicate { it.type.isSolidCollision }).findAny().isPresent || movingCollision(poly)) {
				if (
					// check if we can't pass through because something blocks our path
					world.chunkMap.anyCellSatisfies(poly.aabb) { _, _, cell -> cell.foreground.material.value.collisionKind.isSolidCollision } &&
					// AND check if we CAN pass through, if DYNAMIC tiles were to be removed
					!world.chunkMap.anyCellSatisfies(poly.aabb) { _, _, cell -> cell.foreground.material.value.collisionKind.isSolidNonDynamic }
				) {
					if (!openDoors(poly.aabb)) {
						return false
					}
				} else {
					return false
				}
			}
		}

		return when (edge.action) {
			PathFinder.Action.WALK -> onGround(edge.source.position, Predicate { it.isFloorCollision })
			PathFinder.Action.DROP -> onGround(edge.target.position, Predicate { it.isFloorCollision }) && !onGround(edge.target.position, Predicate { it.isSolidCollision })
			PathFinder.Action.SWIM -> inLiquid(edge.target.position)
			PathFinder.Action.LAND -> onGround(edge.target.position, Predicate { it.isFloorCollision }) || inLiquid(edge.target.position)
			else -> true
		}
	}

	fun move(parameters: ActorMovementParameters, modifiers: ActorMovementModifiers, run: Boolean, dt: Double): Boolean? {
		// pathfind to a new target position in the background while moving on the current path
		updatePathFinding()

		if (path.isEmpty())
			return null

		controlFace = null

		while (path.isNotEmpty()) {
			val edge = path.first()
			val delta = world.geometry.diff(edge.target.position, edge.source.position)

			var sourceVelocity: Vector2d = Vector2d.ZERO
			var targetVelocity: Vector2d = Vector2d.ZERO

			when (edge.action) {
				PathFinder.Action.WALK -> {
					sourceVelocity = delta.unitVector * (if (run) parameters.runSpeed ?: 0.0 else parameters.walkSpeed ?: 0.0) * modifiers.speedModifier
					targetVelocity = sourceVelocity
				}

				PathFinder.Action.JUMP -> {
					if (modifiers.jumpingSuppressed) {
						reset()
						return null
					}
				}

				PathFinder.Action.ARC -> {
					sourceVelocity = edge.source.velocity ?: Vector2d.ZERO
					targetVelocity = edge.target.velocity ?: Vector2d.ZERO
				}

				PathFinder.Action.DROP -> {
					targetVelocity = edge.target.velocity ?: Vector2d.ZERO
				}

				PathFinder.Action.SWIM -> {
					targetVelocity = delta.unitVector * (parameters.flySpeed ?: 0.0) * (1.0 - (parameters.liquidImpedance ?: 0.0))
					sourceVelocity = targetVelocity
				}

				PathFinder.Action.FLY -> {
					// accelerate along path using airForce
					val angleFactor = (controller.velocity.unitVector * delta.unitVector).toAngle()
					val speedAlongAngle = angleFactor * controller.velocity.length
					var acc = (parameters.airForce ?: 0.0) / controller.mass

					if (acc.isInfinite() || acc.isNaN())
						acc = 1.0

					sourceVelocity = delta.unitVector * min(parameters.flySpeed ?: 0.0, speedAlongAngle + acc * dt)
					targetVelocity = sourceVelocity
				}

				PathFinder.Action.LAND -> {}
			}

			val avgVelocity = (sourceVelocity + targetVelocity) / 2.0
			val avgSpeed = avgVelocity.length
			val edgeTime = if (avgSpeed > 0.0) delta.length / avgSpeed else 0.2
			val edgeProgress = edgeTimer / edgeTime

			// original: if (edgeProgress > 1.0f)
			if (edgeProgress >= 1.0) {
				edgeTimer -= edgeTime
				path.removeFirst()

				if (path.isNotEmpty() && !validateEdge(path.first())) {
					reset()
					return null
				}

				continue
			}

			controller.velocity = sourceVelocity + (targetVelocity - sourceVelocity) * edgeProgress
			controller.position = edge.source.position + (controller.velocity + sourceVelocity) / 2.0 * edgeTimer

			var gravity = world.chunkMap.gravityAt(controller.position)

			if (gravity.lengthSquared == 0.0) {
				// zero-g, assume left/right as gravity was pointing straight downwards
				gravity = Vector2d.POSITIVE_Y
			}

			// rotate to right, dot product with delta
			val dot = gravity.rotate(PI / -2.0).dot(delta)

			if (dot > 0.0) {
				controlFace = Direction.RIGHT
			} else if (dot < 0.0) {
				controlFace = Direction.LEFT
			}

			edgeTimer += dt
			return null
		}

		// reached the end of the path, success unless we're also currently pathfinding to a new position
		if (pathFinder != null)
			return null

		return true
	}

	val isPathfinding: Boolean
		get() = path.isEmpty()
}
